// code by jph
package mid.rec;

import javax.sound.midi.MidiMessage;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;

import mid.MidiConstants;

public abstract class EventReceiver implements Receiver {
  @Override
  public final void send(MidiMessage myMidiMessage, long myLong) {
    // typical values of myLong are 11602000, 12370000, ...
    if (myLong % 1000 != 0)
      throw new RuntimeException("better than milliseconds resolution not supported.");
    // ---
    myLong /= 1000; // convert from micro to milli seconds. (the LPK25 has milliseconds resolution.)
    // ---
    byte[] myByte = myMidiMessage.getMessage();
    switch (myByte[0]) {
    case (byte) ShortMessage.NOTE_ON:
      noteOn(myByte[1], myByte[2], myLong);
      break;
    case (byte) ShortMessage.NOTE_OFF:
      // for LPK25 the NOTE_OFF velocity in myByte[2] is always 127!
      noteOff(myByte[1], myLong);
      break;
    case (byte) ShortMessage.CONTROL_CHANGE:
      switch (myByte[1]) {
      case MidiConstants.SUSTAIN_PEDAL:
        // Engages the Sustain Resonance, if it is active, and holds the sustain segment of the envelopes of sounding notes.
        sustain(myByte[2] != 0, myByte[2], myLong);
        break;
      case MidiConstants.SOSTENUTO_PEDAL:
        // Functions as on a real acoustic grand piano, sustaining only on those notes that were depressed when the pedal went down.
        break;
      case MidiConstants.SOFT_PEDAL:
        // Engages the soft pedal samples for that Keyset, if they are active.
        break;
      default:
        reportUnknownMessage(myByte);
      }
      break;
    default:
      reportUnknownMessage(myByte);
    }
  }

  private static void reportUnknownMessage(byte[] myByte) {
    StringBuilder myStringBuffer = new StringBuilder("unexpected midi command:");
    for (int c0 = 0; c0 < myByte.length; ++c0)
      myStringBuffer.append(String.format(" %02x", myByte[c0]));
    throw new RuntimeException(myStringBuffer.toString());
  }

  @Override
  public final void close() {
  }

  /** @param key
   * @param velocity
   * @param myLong timestamp of event in milli seconds after opening device. */
  public abstract void noteOn(int key, int velocity, long myLong);

  /** @param key
   * @param myLong timestamp of event in milli seconds after opening device. */
  public abstract void noteOff(int key, long myLong);

  /** @param sustain 0 for lifted state, != 0 for pressed state
   * @param myLong timestamp of event in milli seconds after opening device. */
  public abstract void sustain(boolean sustain, int myInt, long myLong);
}
