// adapted from Robert Sedgewick and Kevin Wayne
// code by jph
package sys.mat;

/** Data type for complex numbers.
 * 
 * a = 5.0 + 6.0i
 * b = -3.0 + 4.0i
 * Re(a) = 5.0
 * Im(a) = 6.0
 * b + a = 2.0 + 10.0i
 * a - b = 8.0 + 2.0i
 * a * b = -39.0 + 2.0i
 * b * a = -39.0 + 2.0i
 * a / b = 0.36 - 1.52i
 * (a / b) * b = 5.0 + 6.0i
 * conj(a) = 5.0 - 6.0i
 * |a| = 7.810249675906654
 * tan(a) = -6.685231390246571E-6 + 1.0000103108981198i */
public class Complex {
  /** real part */
  public final double re;
  /** imaginary part */
  public final double im;

  /** create a new object with the given real and imaginary parts
   * 
   * @param real
   * @param imag */
  public Complex(double re, double im) {
    this.re = re;
    this.im = im;
  }
  
  public static Complex unit(double phase) {
    return new Complex(0, phase).exp();
  }

  // return abs/modulus/magnitude
  public double abs() {
    return Math.hypot(re, im);
  }

  /** @return angle/phase/argument */
  public double phase() {
    return Math.atan2(im, re);
  } // between -pi and pi

  // return a new Complex object whose value is (this + b)
  public Complex add(Complex b) {
    Complex a = this; // invoking object
    return new Complex(a.re + b.re, a.im + b.im);
  }

  // return a new Complex object whose value is (this - b)
  public Complex sub(Complex b) {
    Complex a = this;
    return new Complex(a.re - b.re, a.im - b.im);
  }

  // return a new Complex object whose value is (this * b)
  public Complex mul(Complex b) {
    Complex a = this;
    return new Complex(a.re * b.re - a.im * b.im, a.re * b.im + a.im * b.re);
  }

  // scalar multiplication
  // return a new object whose value is (this * alpha)
  public Complex mul(double alpha) {
    return new Complex(alpha * re, alpha * im);
  }

  public Complex div(Complex b) {
    return mul(b.reciprocal());
  }

  /** return a new Complex object whose value is the reciprocal of this
   * 
   * @return */
  public Complex reciprocal() {
    double scale = 1 / (re * re + im * im);
    return new Complex(re * scale, -im * scale);
  }

  // return a new Complex object whose value is the conjugate of this
  public Complex conjugate() {
    return new Complex(re, -im);
  }

  // return a new Complex object whose value is the complex exponential of this
  public Complex exp() {
    double exp = Math.exp(re);
    return new Complex(exp * Math.cos(im), exp * Math.sin(im));
  }

  // return a new Complex object whose value is the complex sine of this
  public Complex sin() {
    return new Complex(Math.sin(re) * Math.cosh(im), Math.cos(re) * Math.sinh(im));
  }

  // return a new Complex object whose value is the complex cosine of this
  public Complex cos() {
    return new Complex(Math.cos(re) * Math.cosh(im), -Math.sin(re) * Math.sinh(im));
  }

  // return a new Complex object whose value is the complex tangent of this
  public Complex tan() {
    return sin().div(cos());
  }

  @Override
  public String toString() {
    if (im == 0)
      return re + "";
    if (re == 0)
      return im + "i";
    if (im < 0)
      return re + " - " + (-im) + "i";
    return re + " + " + im + "i";
  }

  // sample client for testing
  public static void main(String[] args) {
    Complex a = new Complex(5.0, 6.0);
    Complex b = new Complex(-3.0, 4.0);
    System.out.println("a            = " + a);
    System.out.println("b            = " + b);
    System.out.println("b + a        = " + b.add(a));
    System.out.println("a - b        = " + a.sub(b));
    System.out.println("a * b        = " + a.mul(b));
    System.out.println("b * a        = " + b.mul(a));
    System.out.println("a / b        = " + a.div(b));
    System.out.println("(a / b) * b  = " + a.div(b).mul(b));
    System.out.println("conj(a)      = " + a.conjugate());
    System.out.println("|a|          = " + a.abs());
    System.out.println("tan(a)       = " + a.tan());
  }
}
