// code by jph
package sys.mat;

/** integer fraction in normal form non-degenerate, i.e. denominator != 0 immutable */
public class Fraction extends Number implements Comparable<Fraction> {
  public static final Fraction zero = new Fraction(0, 1);
  public static final Fraction one = new Fraction(1, 1);
  // ---
  /** numerator */
  public final int num;
  /** denominator (always greater than zero) */
  public final int den;

  /** @param num
   * @param den has to be non-zero
   * @throws {@link ArithmeticException} if den is zero */
  public Fraction(final int num, final int den) {
    if (den == 0)
      throw new ArithmeticException(num + "/" + den);
    int gcd = IntegerMath.gcd(num, den);
    int res = den / gcd;
    if (0 < res) {
      this.num = num / gcd;
      this.den = res;
    } else {
      this.num = -num / gcd;
      this.den = -res;
    }
  }

  public static Fraction reciprocal(int den) {
    return new Fraction(1, den);
  }

  /** separator character is '/' whitespace is removed before Integer.parseInt is invoked
   * 
   * @param myString
   * @return */
  public static Fraction parseFraction(String myString) {
    String[] mySplit = myString.split("/");
    return new Fraction(Integer.parseInt(mySplit[0].trim()), Integer.parseInt(mySplit[1].trim()));
  }

  public Fraction negate() {
    return new Fraction(-num, den);
  }

  /** uses gcd of the denominators (better only in special cases than the straight forward formula)
   * 
   * @param myFraction
   * @return */
  public Fraction add(Fraction myFraction) {
    final int gcd = IntegerMath.gcd(den, myFraction.den);
    final int red = myFraction.den / gcd;
    return new Fraction(num * red + myFraction.num * (den / gcd), den * red);
  }

  public Fraction sub(Fraction myFraction) {
    return add(myFraction.negate());
  }

  /** multiplication using crossover of factors to minimize risk of overflow
   * 
   * @param myFraction
   * @return */
  public Fraction mul(Fraction myFraction) {
    return new Fraction(num, myFraction.den).private_mul(new Fraction(myFraction.num, den));
  }

  public Fraction times(int myInt) {
    return mul(new Fraction(myInt, 1));
  }

  private final Fraction private_mul(Fraction myFraction) {
    return new Fraction(num * myFraction.num, den * myFraction.den);
  }

  public Fraction div(Fraction myFraction) {
    return mul(myFraction.inverse());
  }

  /** also denoted reciprocal
   * 
   * @return den/num */
  public final Fraction inverse() {
    return new Fraction(den, num);
  }

  /** @return fraction between 0<= and <1 */
  public Fraction getFractional() {
    return new Fraction(IntegerMath.mod(num, den), den);
  }

  public int floor() {
    return IntegerMath.floorDiv(num, den);
  }

  public int ceil() {
    return IntegerMath.ceilDiv(num, den);
  }

  /** adds 1/2 to the fraction, and floors result
   * 
   * @return */
  public int round() {
    return add(new Fraction(1, 2)).floor(); // is (num + den/2, den) sufficient?
  }

  @Override
  public int intValue() {
    return round();
  }

  @Override
  public long longValue() {
    return round();
  }

  @Override
  public float floatValue() {
    return (float) doubleValue();
  }

  @Override
  public double doubleValue() {
    return num / (double) den;
  }

  /** @return true if den == 1 */
  public boolean isInteger() {
    return den == 1;
  }

  /** call only if isInteger() returns true otherwise an exception is thrown
   * 
   * @return num */
  public int toInteger() {
    if (!isInteger()) // assert den == 1
      throw new RuntimeException(this + " is no integer");
    return num;
  }

  /** @return -18+4/7 for display, not parsing */
  public String toHighSchoolString() {
    final int floor = floor();
    if (floor == 0)
      return isInteger() ? "" + num : toString();
    StringBuilder myStringBuilder = new StringBuilder();
    myStringBuilder.append(floor);
    Fraction myFractional = getFractional();
    if (!myFractional.equals(zero))
      myStringBuilder.append("+" + myFractional);
    return myStringBuilder.toString();
  }

  @Override
  public int compareTo(Fraction myFraction) {
    // return num * myFraction.den - den * myFraction.num; // does not work for big integers
    return Double.compare(doubleValue(), myFraction.doubleValue());
  }

  public boolean equalsInteger(int value) {
    return equals(new Fraction(value, 1));
  }

  @Override
  public boolean equals(Object myObject) {
    if (myObject == null) // || !(myObject instanceof Fraction)
      return false;
    Fraction myFraction = (Fraction) myObject;
    return num == myFraction.num && den == myFraction.den; // sufficient since in normal form
  }

  @Override
  public int hashCode() {
    // n / 1 is mapped to n, thus is identical to getInteger()
    return num + den - 1;
  }

  @Override
  public String toString() {
    return num + "/" + den;
  }
}
