// code by jph
// http://stackoverflow.com/questions/3305059/how-do-you-calculate-log-base-2-in-java-for-integers
package sys.mat;

import java.util.Collection;

public class IntegerMath {
  private IntegerMath() {
  }

  public static int log2(int n) {
    if (n <= 0)
      throw new IllegalArgumentException();
    return 31 - Integer.numberOfLeadingZeros(n);
  }

  public static boolean isPowerOf2(int n) {
    return 0 < n ? (n & (n - 1)) == 0 : false;
  }

  /** mod that behaves like in Matlab. for instance mod(-10, 3) == 2
   * 
   * @param index
   * @param size
   * @return matlab.mod(index, size) */
  public static int mod(int index, int size) {
    int value = index % size;
    // if value is below 0, then -size < value && value < 0.
    // For instance: -3%3==0, and -2%3==-2.
    return value < 0 ? size + value : value;
  }

  /** Euclid's algorithm
   * 
   * @param a
   * @param b
   * @return greatest common divider of a and b. */
  public static int gcd(int a, int b) {
    return b == 0 ? a : gcd(b, a % b);
  }

  public static int lcm(int a, int b) {
    return a * (b / gcd(a, b)); // to avoid overflow
  }

  /** @param myCollection non-empty
   * @return greatest common divider of all integers in myCollection */
  public static int gcd(Collection<Integer> myCollection) {
    return myCollection.stream().reduce(IntegerMath::gcd).orElse(null);
  }

  /** @param myCollection non-empty
   * @return least common multiple of all integers in myCollection */
  public static int lcm(Collection<Integer> myCollection) {
    return myCollection.stream().reduce(IntegerMath::lcm).orElse(null);
  }

  public static int floorDiv(int a, int b) {
    return 0 <= a ? a / b : (a - b + 1) / b;
  }

  public static int ceilDiv(int a, int b) {
    return floorDiv(a + b - 1, b);
  }

  /** method is faster than (int)Math.floor(x)
   * 
   * @param x
   * @return */
  public static int floor(double x) {
    int xi = (int) x;
    return x < xi ? xi - 1 : xi;
  }

  /** @param ofs
   * @param mod
   * @param min
   * @return minimal s with s ==_mod ofs && min <= s */
  public static int findMin(int ofs, int mod, int min) {
    int value = ofs - floorDiv(ofs - min, mod) * mod;
    if (value < min || min + mod <= value)
      throw new RuntimeException();
    return value;
  }

  /** @param ofs
   * @param mod
   * @param min
   * @return maximal s with s ==_mod ofs && s <= max */
  public static int findMax(int ofs, int mod, int max) {
    int value = ofs + floorDiv(max - ofs, mod) * mod;
    if (max < value || value + mod <= max)
      throw new RuntimeException();
    return value;
  }

  public static void main(String[] args) {
    int ofs = 3;
    int mod = 7;
    for (int count = -99; count < 100; ++count) {
      int beg = findMax(ofs, mod, count);
      int end = findMin(ofs, mod, count);
      System.out.println(String.format("max %4d <= %4d  %d]   [min %4d <= %4d  %d", //
          beg, count, (beg - ofs) % mod, //
          count, end, (end - ofs) % mod));
    }
  }
}
