// code by jph
package sys.mat.num;

import java.util.Random;

public abstract class BasisSpline {
  public static final BasisSpline constant = new BasisSpline() {
    @Override
    public double at(double t) {
      return -0.5 <= t && t < +0.5 ? 1 : 0;
    }

    @Override
    public int support() {
      return 1;
    }
  };
  public static final BasisSpline linear = new BasisSpline() {
    @Override
    public double at(double t) {
      return Math.max(0, 1 - Math.abs(t));
    }

    @Override
    public int support() {
      return 2;
    }
  };
  public static final BasisSpline quadratic = new BasisSpline() {
    @Override
    public double at(double t) {
      if (t < 0)
        return at(-t);
      double t2 = t * t;
      if (t < +0.5)
        return 0.75 - t2;
      if (+0.5 <= t && t < +1.5)
        return 1.125 - 1.5 * t + 0.5 * t2;
      return 0;
    }

    @Override
    public int support() {
      return 3;
    }
  };
  public static final BasisSpline cubic = new BasisSpline() {
    @Override
    public double at(double t) {
      if (t < 0)
        return at(-t);
      if (t < 1) {
        double t2 = t * t;
        return 2 / 3. - t2 + 0.5 * t2 * t;
      }
      if (t < 2) {
        double t2 = 2 - t;
        return 1 / 6. * t2 * t2 * t2;
      }
      return 0;
    }

    @Override
    public int support() {
      return 4;
    }
  };
  public static final BasisSpline[] of = new BasisSpline[] { constant, linear, quadratic, cubic };

  public abstract double at(double t);

  public abstract int support();

  private double unity(double t) {
    t -= Math.floor(t);
    double s = 0;
    int r = support();
    for (int c0 = -r; c0 <= r; ++c0)
      s += at(t + c0);
    return s;
  }

  public static void main(String[] args) {
    BasisSpline myBasisSpline = BasisSpline.quadratic;
    System.out.println(myBasisSpline.support());
    Random myRandom = new Random(System.currentTimeMillis());
    for (int c = 0; c < 10; ++c) {
      double val = myRandom.nextDouble();
      double ext = val * myBasisSpline.support() - myBasisSpline.support() / 2.;
      System.out.println(String.format("%8.5f ",ext) + " " + myBasisSpline.unity(ext));
    }
  }
}
